"""Loading and validation of config files."""

import datetime
from typing import (
    Any,
    Dict,
    List,
    Union,
    Mapping,
    Pattern,
    Iterable,
    Sequence,
    NamedTuple,
)

from dataclasses import dataclass

from routemaster.exit_conditions import ExitConditionProgram

if False:  # typing
    from routemaster.context import Context  # noqa


class TimeTrigger(NamedTuple):
    """Time based trigger for exit condition evaluation."""
    time: datetime.time


class IntervalTrigger(NamedTuple):
    """Time based trigger for exit condition evaluation."""
    interval: datetime.timedelta


class MetadataTrigger(NamedTuple):
    """Context update based trigger for exit condition evaluation."""
    metadata_path: str

    def should_trigger_for_update(self, update: Dict[str, Any]) -> bool:
        """Returns whether this trigger should fire for a given update."""
        def applies(path: Sequence[str], d: Dict[str, Any]) -> bool:
            component, path = path[0], path[1:]
            if component in d:
                if path:
                    return applies(path, d[component])
                return True
            return False
        return applies(self.metadata_path.split('.'), update)


@dataclass
class OnEntryTrigger:
    """Trigger on entry to a given gate."""


Trigger = Union[TimeTrigger, IntervalTrigger, MetadataTrigger, OnEntryTrigger]


class ConstantNextState(NamedTuple):
    """Defines a constant choice, always chooses `state`."""
    state: str

    def next_state_for_label(self, label_context: Any) -> str:
        """Returns the constant next state."""
        return self.state

    def all_destinations(self) -> Iterable[str]:
        """Returns the constant next state."""
        return [self.state]


class ContextNextStatesOption(NamedTuple):
    """Represents an option for a context conditional next state."""
    state: str
    value: Any


class ContextNextStates(NamedTuple):
    """Defined a choice based on a path in the given `label_context`."""
    path: str
    destinations: Iterable[ContextNextStatesOption]
    default: str  # noqa: E704 misidentified "multiple statements on one line"

    def next_state_for_label(self, label_context: 'Context') -> str:
        """Returns next state based on context value at `self.path`."""
        val = label_context.lookup(self.path.split('.'))
        for destination in self.destinations:
            if destination.value == val:
                return destination.state
        return self.default

    def all_destinations(self) -> Iterable[str]:
        """Returns all possible destination states."""
        return [x.state for x in self.destinations] + [self.default]


class NoNextStates(NamedTuple):
    """Represents the lack of a next state to progress to."""

    def next_state_for_label(self, label_context: Any) -> str:
        """Invalid to call, raise an exception."""
        raise RuntimeError(
            "Attempted to progress from a state with no next state",
        )

    def all_destinations(self) -> Iterable[str]:
        """Returns no states."""
        return []


NextStates = Union[ConstantNextState, ContextNextStates, NoNextStates]


class Gate(NamedTuple):
    """
    A state that restricts a label from moving based on an exit condition.

    Gates cannot perform an action.
    """
    name: str
    next_states: NextStates

    exit_condition: ExitConditionProgram
    triggers: Iterable[Trigger]

    @property
    def metadata_triggers(self) -> List[MetadataTrigger]:
        """Return a list of the metadata triggers for this state."""
        return [x for x in self.triggers if isinstance(x, MetadataTrigger)]

    @property
    def trigger_on_entry(self) -> bool:
        """Util to check if this gate should be triggered on entry."""
        return any(isinstance(x, OnEntryTrigger) for x in self.triggers)


class Action(NamedTuple):
    """
    A state that performs an action via a webhook.

    A label staying in this state means that the action has not succeeded, i.e.
    the webhook returned an error status.
    """
    name: str
    next_states: NextStates

    webhook: str


State = Union[Action, Gate]


class FeedConfig(NamedTuple):
    """
    The definition of a feed of dynamic data to be included in a context.
    """
    name: str
    url: str


class Webhook(NamedTuple):
    """Configuration for webdook requests."""
    match: Pattern
    headers: Dict[str, str]


class StateMachine(NamedTuple):
    """A state machine."""
    name: str
    states: List[State]
    feeds: List[FeedConfig]
    webhooks: List[Webhook]

    def get_state(self, state_name: str) -> State:
        """Get the state object for a given state name."""
        return [x for x in self.states if x.name == state_name][0]


class DatabaseConfig(NamedTuple):
    """Database connection configuration."""
    host: str
    port: int
    name: str
    username: str
    password: str

    @property
    def connstr(self) -> str:
        """Connection string for given configuration."""
        if not self.host:
            return f'postgresql:///{self.name}'

        auth = ''
        if self.username and not self.password:
            auth = f'{self.username}@'
        elif self.username and self.password:
            auth = f'{self.username}:{self.password}@'

        return f'postgresql://{auth}{self.host}:{self.port}/{self.name}'


class Config(NamedTuple):
    """
    The top-level configuration object.

    Stores the configured state machines, and other system-level configuration.
    """
    state_machines: Mapping[str, StateMachine]
    database: DatabaseConfig



"""Database model definition."""
import datetime
import functools

import dateutil.tz
from sqlalchemy import DDL, Table
from sqlalchemy import Column as NullableColumn
from sqlalchemy import (
    String,
    Boolean,
    Integer,
    DateTime,
    MetaData,
    FetchedValue,
    ForeignKeyConstraint,
    func,
)
from sqlalchemy.dialects.postgresql import JSONB

metadata = MetaData()

Column = functools.partial(NullableColumn, nullable=False)

sync_label_updated_column = DDL(
    '''
    CREATE OR REPLACE FUNCTION sync_label_updated_column_fn()
        RETURNS TRIGGER AS
            $$
                BEGIN
                    NEW.updated = now() AT TIME ZONE 'UTC';
                    RETURN NEW;
                END;
            $$
        LANGUAGE PLPGSQL;

    CREATE TRIGGER sync_label_updated_column
        BEFORE UPDATE ON labels
        FOR EACH ROW
        EXECUTE PROCEDURE sync_label_updated_column_fn();
    ''',
)


"""The representation of the state of a label."""
labels = Table(
    'labels',
    metadata,
    Column('name', String, primary_key=True),
    Column('state_machine', String, primary_key=True),
    Column('metadata', JSONB),
    Column('metadata_triggers_processed', Boolean, default=True),
    Column('deleted', Boolean, default=False),
    Column(
        'updated',
        DateTime(timezone=True),
        server_default=func.now(),
        server_onupdate=FetchedValue(),
    ),
    listeners=[
        ('after_create', sync_label_updated_column),
    ],
)


"""Represents history of state transitions for a label."""
history = Table(
    'history',
    metadata,
    Column('id', Integer, primary_key=True, autoincrement=True),

    Column('label_name', String),
    Column('label_state_machine', String),
    ForeignKeyConstraint(
        ['label_name', 'label_state_machine'],
        ['labels.name', 'labels.state_machine'],
    ),

    Column(
        'created',
        DateTime(timezone=True),
        default=lambda: datetime.datetime.now(dateutil.tz.tzutc()),
    ),

    # `forced = True` represents a manual transition that may not be in
    # accordance with the state machine logic.
    Column('forced', Boolean, default=False),

    # Null indicates starting a state machine
    NullableColumn('old_state', String),

    # Null indicates being deleted from a state machine
    NullableColumn('new_state', String),
)



"""Processing of transitions between states."""

import logging
import textwrap

from routemaster.app import App
from routemaster.config import Gate, Action
from routemaster.state_machine.gates import process_gate
from routemaster.state_machine.types import LabelRef
from routemaster.state_machine.utils import (
    lock_label,
    get_current_state,
    get_state_machine,
)
from routemaster.state_machine.actions import process_action
from routemaster.state_machine.exceptions import DeletedLabel

logger = logging.getLogger(__name__)

# The maximum transitions that may happen in a single `process_transitions`.
# Note that later transitions above this number will still happen, but after
# yielding to other cron processes, etc.
MAX_TRANSITIONS = 50


def process_transitions(app: App, label: LabelRef) -> None:
    """
    Process each transition for a label until it cannot move any further.

    Will silently accept DeletedLabel exceptions and end the processing of
    transitions.
    """

    state_machine = get_state_machine(app, label)
    could_progress = True
    num_transitions = 0

    def _transition() -> bool:
        with app.db.begin() as conn:
            lock_label(label, conn)
            current_state = get_current_state(label, state_machine, conn)

            if isinstance(current_state, Action):
                return process_action(
                    app=app,
                    state=current_state,
                    state_machine=state_machine,
                    label=label,
                    conn=conn,
                )

            elif isinstance(current_state, Gate):  # pragma: no branch
                if not current_state.trigger_on_entry:
                    return False

                return process_gate(
                    app=app,
                    state=current_state,
                    state_machine=state_machine,
                    label=label,
                    conn=conn,
                )

            else:
                raise RuntimeError(  # pragma: no cover
                    "Unsupported state type {0}".format(current_state),
                )

    while could_progress and num_transitions < MAX_TRANSITIONS:
        num_transitions += 1

        if num_transitions == MAX_TRANSITIONS:
            logger.warn(textwrap.dedent(
                f"""
                Label {label}
                hit the maximum number of transitions allowed in one go. This
                may indicate a bug, or could be negatively impacting your
                Routemaster cron processing. If it's not a bug, try using gates
                with time based exit conditions to break up the processing, or
                submit an issue or pull request to
                https://github.com/thread/routemaster with your use-case.
                """,
            ))

        try:
            could_progress = _transition()
        except DeletedLabel:
            # Label might have been deleted, that's a supported use-case,
            # not even a warning, and we should allow the this process to
            # continue.
            return



"""Exceptions for use in exit condition handling."""


class ParseError(Exception):
    """Errors that occur when tokenizing or parsing."""

    def __init__(self, message, location):
        """
        Construct by message and location.

        location is given as a 2-tuple of a start and end position in the
        source.
        """
        self.message = message
        self.location = location



"""Top-level utility for exit condition programs."""

import typing

from routemaster.exit_conditions.parser import parse
from routemaster.exit_conditions.analysis import find_accessed_keys
from routemaster.exit_conditions.peephole import peephole_optimise
from routemaster.exit_conditions.evaluator import evaluate
from routemaster.exit_conditions.exceptions import ParseError
from routemaster.exit_conditions.error_display import (
    format_parse_error_message,
)

if False:  # typing
    from routemaster.context import Context  # noqa


class ExitConditionProgram(object):
    """Compiled exit condition program."""

    def __init__(self, source: str) -> None:
        """
        Construct from source.

        This will eagerly compile and report any errors.
        """
        try:
            self._instructions = tuple(parse(source))
        except ParseError as exc:
            raise ValueError(format_parse_error_message(
                source=source,
                error=exc,
            )) from None

        self._instructions = tuple(peephole_optimise(self._instructions))

        self.source = source

    def accessed_variables(self) -> typing.Iterable[str]:
        """Iterable of names of variables accessed in this program."""
        for accessed_key in find_accessed_keys(self._instructions):
            yield '.'.join(accessed_key)

    def run(self, context: 'Context') -> bool:
        """Evaluate this program with a given context."""
        return evaluate(
            self._instructions,
            context.lookup,
            context.property_handler,
        )

    def __eq__(self, other_program: typing.Any) -> bool:
        """
        Equality test.

        Programs are compared as equal iff their instruction sequences compare
        as equal. This means that the equivalence relation will treat, for
        instance, `foo = bar` and `(foo    = bar) # hats` as equal, but not
        `bar = foo` and `foo = bar`.
        """
        if not isinstance(other_program, ExitConditionProgram):
            return False

        return other_program._instructions == self._instructions

    def __hash__(self) -> int:
        """
        Identity hashing.

        This preserves the equivalence relation from `__eq__`: that is, if
        a == b for programs a, b then hash(a) == hash(b).
        """
        return hash(tuple(self._instructions))

    def __repr__(self) -> str:
        """Convenient debug representation."""
        return f"{type(self).__name__}({self.source!r})"



"""Data model for configuration format."""

import os
import re
import datetime
from typing import Any, Dict, List, Optional

import yaml
import jsonschema
import pkg_resources
import jsonschema.exceptions

from routemaster.config.model import (
    Gate,
    State,
    Action,
    Config,
    Trigger,
    Webhook,
    FeedConfig,
    NextStates,
    TimeTrigger,
    NoNextStates,
    StateMachine,
    DatabaseConfig,
    OnEntryTrigger,
    IntervalTrigger,
    MetadataTrigger,
    ConstantNextState,
    ContextNextStates,
    ContextNextStatesOption,
)
from routemaster.exit_conditions import ExitConditionProgram
from routemaster.config.exceptions import ConfigError

Yaml = Dict[str, Any]
Path = List[str]


def load_config(yaml: Yaml) -> Config:
    """Unpack a parsed YAML file into a `Config` object."""
    _schema_validate(yaml)

    try:
        yaml_state_machines = yaml['state_machines']
    except KeyError:  # pragma: no cover
        raise ConfigError(
            "No top-level state_machines key defined.",
        ) from None

    return Config(
        state_machines={
            name: _load_state_machine(
                ['state_machines', name],
                name,
                yaml_state_machine,
            )
            for name, yaml_state_machine in yaml_state_machines.items()
        },
        database=load_database_config(),
    )


def _schema_validate(config: Yaml) -> None:
    # Load schema from package resources
    schema_raw = pkg_resources.resource_string(
        'routemaster.config',
        'schema.yaml',
    ).decode('utf-8')
    schema_yaml = yaml.load(schema_raw)

    try:
        jsonschema.validate(config, schema_yaml)
    except jsonschema.exceptions.ValidationError:
        raise ConfigError("Could not validate config file against schema.")


def load_database_config() -> DatabaseConfig:
    """Load the database config from the environment."""
    port_string = os.environ.get('DB_PORT', 5432)

    try:
        port = int(port_string)
    except ValueError:
        raise ConfigError(
            f"Could not parse DB_PORT as an integer: '{port_string}'.",
        )

    return DatabaseConfig(
        host=os.environ.get('DB_HOST', 'localhost'),
        port=port,
        name=os.environ.get('DB_NAME', 'routemaster'),
        username=os.environ.get('DB_USER', 'routemaster'),
        password=os.environ.get('DB_PASS', ''),
    )


def _load_state_machine(
    path: Path,
    name: str,
    yaml_state_machine: Yaml,
) -> StateMachine:
    feeds = [_load_feed_config(x) for x in yaml_state_machine.get('feeds', [])]

    if len(set(x.name for x in feeds)) < len(feeds):
        raise ConfigError(
            f"FeedConfigs must have unique names at "
            f"{'.'.join(path + ['feeds'])}",
        )

    return StateMachine(
        name=name,
        states=[
            _load_state(path + ['states', str(idx)], yaml_state)
            for idx, yaml_state in enumerate(yaml_state_machine['states'])
        ],
        feeds=feeds,
        webhooks=[
            _load_webhook(x)
            for x in yaml_state_machine.get('webhooks', [])
        ],
    )


def _load_webhook(yaml: Yaml) -> Webhook:
    return Webhook(
        match=re.compile(yaml['match']),
        headers=yaml['headers'],
    )


def _load_feed_config(yaml: Yaml) -> FeedConfig:
    return FeedConfig(name=yaml['name'], url=yaml['url'])


def _load_state(path: Path, yaml_state: Yaml) -> State:
    if 'action' in yaml_state and 'gate' in yaml_state:  # pragma: no branch
        raise ConfigError(  # pragma: no cover
            f"State at path {'.'.join(path)} cannot be both a gate and an "
            f"action.",
        )

    if 'action' in yaml_state:
        return _load_action(path, yaml_state)
    elif 'gate' in yaml_state:  # pragma: no branch
        return _load_gate(path, yaml_state)
    else:
        raise ConfigError(  # pragma: no cover
            f"State at path {'.'.join(path)} must be either a gate or an "
            f"action.",
        )


def _load_action(path: Path, yaml_state: Yaml) -> Action:
    return Action(
        name=yaml_state['action'],
        webhook=yaml_state['webhook'],
        next_states=_load_next_states(
            path + ['next'],
            yaml_state.get('next'),
        ),
    )


def _load_gate(path: Path, yaml_state: Yaml) -> Gate:
    yaml_exit_condition = yaml_state['exit_condition']

    if yaml_exit_condition is True:
        str_exit_condition = 'true'
    elif yaml_exit_condition is False:
        str_exit_condition = 'false'
    else:
        str_exit_condition = str(yaml_exit_condition).strip()

    return Gate(
        name=yaml_state['gate'],
        exit_condition=ExitConditionProgram(str_exit_condition),
        triggers=[
            _load_trigger(path + ['triggers', str(idx)], yaml_trigger)
            for idx, yaml_trigger in enumerate(yaml_state.get('triggers', []))
        ],
        next_states=_load_next_states(path + ['next'], yaml_state.get('next')),
    )


def _load_trigger(path: Path, yaml_trigger: Yaml) -> Trigger:
    if len(yaml_trigger.keys()) > 1:  # pragma: no branch
        raise ConfigError(  # pragma: no cover
            f"Trigger at path {'.'.join(path)} cannot be of multiple types.",
        )

    if 'time' in yaml_trigger:
        return _load_time_trigger(path, yaml_trigger)
    elif 'metadata' in yaml_trigger:
        return _load_metadata_trigger(path, yaml_trigger)
    elif 'interval' in yaml_trigger:  # pragma: no branch
        return _load_interval_trigger(path, yaml_trigger)
    elif yaml_trigger.get('event') == 'entry':
        return OnEntryTrigger()
    else:
        raise ConfigError(  # pragma: no cover
            f"Trigger at path {'.'.join(path)} must be a time, interval, or "
            f"metadata trigger.",
        )


def _load_time_trigger(path: Path, yaml_trigger: Yaml) -> TimeTrigger:
    format_ = '%Hh%Mm'
    try:
        dt = datetime.datetime.strptime(str(yaml_trigger['time']), format_)
        trigger = dt.time()
    except ValueError:  # pragma: no cover
        raise ConfigError(
            f"Time trigger '{yaml_trigger['time']}' at path {'.'.join(path)} "
            f"does not meet expected format: {format_}.",
        ) from None
    return TimeTrigger(time=trigger)


RE_INTERVAL = re.compile(
    r'((?P<days>\d+?)d)?((?P<hours>\d+?)h)?'
    r'((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?',
)


def _load_interval_trigger(path: Path, yaml_trigger: Yaml) -> IntervalTrigger:
    match = RE_INTERVAL.match(yaml_trigger['interval'])
    if not match:  # pragma: no branch
        raise ConfigError(  # pragma: no cover
            f"Interval trigger '{yaml_trigger['interval']}' at path "
            f"{'.'.join(path)} does not meet expected format: 'XdYhZm'.",
        )

    parts = match.groupdict()
    interval = datetime.timedelta(**{
        x: int(y) if y is not None else 0
        for x, y in parts.items()
    })
    return IntervalTrigger(interval=interval)


RE_PATH = re.compile(r'^[a-zA-Z0-9_]+(\.[a-zA-Z0-9_]+)*$')


def _load_metadata_trigger(path: Path, yaml_trigger: Yaml) -> MetadataTrigger:
    metadata_path = yaml_trigger['metadata']
    if not RE_PATH.match(metadata_path):  # pragma: no branch
        raise ConfigError(  # pragma: no cover
            f"Metadata trigger '{metadata_path}' at path {'.'.join(path)} is "
            f"not a valid dotted path.",
        )
    return MetadataTrigger(metadata_path=metadata_path)


def _load_next_states(
    path: Path,
    yaml_next_states: Optional[Yaml],
) -> NextStates:

    if yaml_next_states is None:
        return NoNextStates()

    if isinstance(yaml_next_states, str):
        return _load_constant_next_state(path, {'state': yaml_next_states})
    if yaml_next_states['type'] == 'constant':
        return _load_constant_next_state(path, yaml_next_states)
    elif yaml_next_states['type'] == 'context':  # pragma: no branch
        return _load_context_next_states(path, yaml_next_states)
    else:
        raise ConfigError(  # pragma: no cover
            f"Next state config at path {'.'.join(path)} must be of type "
            f"'constant' or 'context'",
        ) from None


def _load_constant_next_state(
    path: Path,
    yaml_next_states: Yaml,
) -> NextStates:
    return ConstantNextState(state=yaml_next_states['state'])


def _load_context_next_states(
    path: Path,
    yaml_next_states: Yaml,
) -> NextStates:
    return ContextNextStates(
        path=yaml_next_states['path'],
        destinations=[
            _load_context_next_state_option(
                path + ['destinations', str(idx)],
                yaml_option,
            )
            for idx, yaml_option in enumerate(yaml_next_states['destinations'])
        ],
        default=yaml_next_states['default'],
    )


def _load_context_next_state_option(
    path: Path,
    yaml_option: Yaml,
) -> ContextNextStatesOption:
    return ContextNextStatesOption(
        state=yaml_option['state'],
        value=yaml_option['value'],
    )



"""Top-level gunicorn application for `routemaster serve`."""

from typing import Any, Dict, Callable, Iterable

import werkzeug.debug
import gunicorn.app.base

StartResponse = Callable[
    [
        str,
        Dict[str, str],
        Any,
    ],
    None,
]

WSGICallable = Callable[
    [
        Dict[str, Any],
        StartResponse,
    ],
    Iterable[bytes],
]


class GunicornWSGIApplication(gunicorn.app.base.BaseApplication):
    """gunicorn application for routemaster."""

    def __init__(
        self,
        app: WSGICallable,
        *,
        bind: str,
        log_level: str,
        debug: bool,
    ) -> None:
        self.application = app
        self.bind = bind
        self.log_level = log_level
        self.debug = debug
        super().__init__()

    def load_config(self) -> None:
        """
        Load gunicorn configuration settings.

        Rather than grab these from a file we instead just set them to their
        known values inline.
        """
        self.cfg.set('bind', self.bind)
        self.cfg.set('workers', 1)
        self.cfg.set('loglevel', self.log_level)

        if self.debug:
            self.cfg.set('reload', True)
            self.cfg.set('accesslog', '-')

    def load(self) -> WSGICallable:
        """
        Load gunicorn WSGI callable.

        Luckily little loading is needed since this is available inline.
        """
        if self.debug:
            return werkzeug.debug.DebuggedApplication(
                self.application,
                evalex=False,
            )
        return self.application



"""The core of the state machine logic."""

import logging
from typing import Any, Callable, Iterable

from sqlalchemy import and_, not_
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import select

from routemaster.db import labels, history
from routemaster.app import App
from routemaster.utils import dict_merge, suppress_exceptions
from routemaster.config import Gate, State, StateMachine
from routemaster.state_machine.gates import process_gate
from routemaster.state_machine.types import LabelRef, Metadata
from routemaster.state_machine.utils import (
    lock_label,
    get_current_state,
    get_state_machine,
)
from routemaster.state_machine.utils import \
    get_label_metadata as get_label_metadata_internal
from routemaster.state_machine.utils import (
    start_state_machine,
    needs_gate_evaluation_for_metadata_change,
)
from routemaster.state_machine.exceptions import (
    DeletedLabel,
    UnknownLabel,
    LabelAlreadyExists,
)
from routemaster.state_machine.transitions import process_transitions

logger = logging.getLogger(__name__)


def list_labels(app: App, state_machine: StateMachine) -> Iterable[LabelRef]:
    """
    Returns a sorted iterable of labels associated with a state machine.

    Labels are returned ordered alphabetically by name.
    """
    with app.db.begin() as conn:
        label_names = conn.execute(
            select([
                labels.c.name,
            ]).where(and_(
                labels.c.state_machine == state_machine.name,
                not_(labels.c.deleted),
            )).order_by(
                labels.c.name,
            ),
        )
        for row in label_names:
            yield LabelRef(row[labels.c.name], state_machine.name)


def get_label_state(app: App, label: LabelRef) -> State:
    """Finds the current state of a label."""
    state_machine = get_state_machine(app, label)

    with app.db.begin() as conn:
        return get_current_state(label, state_machine, conn)


def get_label_metadata(app: App, label: LabelRef) -> Metadata:
    """Returns the metadata associated with a label."""
    state_machine = get_state_machine(app, label)

    with app.db.begin() as conn:
        row = get_label_metadata_internal(label, state_machine, conn)

        if row is None:
            raise UnknownLabel(label)

        metadata, deleted = row
        if deleted:
            raise DeletedLabel(label)

        return metadata


def create_label(app: App, label: LabelRef, metadata: Metadata) -> Metadata:
    """Creates a label and starts it in a state machine."""
    state_machine = get_state_machine(app, label)

    with app.db.begin() as conn:
        try:
            conn.execute(labels.insert().values(
                name=label.name,
                state_machine=state_machine.name,
                metadata=metadata,
            ))
        except IntegrityError:
            raise LabelAlreadyExists(label)

        start_state_machine(app, label, conn)

    # Outside transaction
    process_transitions(app, label)
    return metadata


def update_metadata_for_label(
    app: App,
    label: LabelRef,
    update: Metadata,
) -> Metadata:
    """
    Updates the metadata for a label.

    Moves the label through the state machine as appropriate.
    """
    state_machine = get_state_machine(app, label)
    needs_gate_evaluation = False

    with app.db.begin() as conn:
        row = lock_label(label, conn)

        existing_metadata, deleted = row.metadata, row.deleted
        if deleted:
            raise DeletedLabel(label)

        needs_gate_evaluation, current_state = \
            needs_gate_evaluation_for_metadata_change(
                state_machine,
                label,
                update,
                conn,
            )

        new_metadata = dict_merge(existing_metadata, update)

        conn.execute(labels.update().where(and_(
            labels.c.name == label.name,
            labels.c.state_machine == label.state_machine,
        )).values(
            metadata=new_metadata,
            metadata_triggers_processed=not needs_gate_evaluation,
        ))

    # Outside transaction
    # Try to move the label forward, but this is not a hard requirement as
    # the cron will come back around to progress the label later.
    if needs_gate_evaluation:
        try:
            _process_transitions_for_metadata_update(
                app,
                label,
                state_machine,
                current_state,
            )
        except Exception:
            # This is allowed to fail here. We have successfully saved the new
            # metadata, and it has a metadata_triggers_processed=False flag so
            # will be picked up again for processing later.
            pass

    return new_metadata


def _process_transitions_for_metadata_update(
    app: App,
    label: LabelRef,
    state_machine: StateMachine,
    state_pending_update: State,
):
    with app.db.begin() as conn:
        lock_label(label, conn)
        current_state = get_current_state(label, state_machine, conn)

        if state_pending_update != current_state:
            # We have raced with another update, and are no longer in
            # the state for which we needed an update, so we should
            # stop.
            return

        if not isinstance(current_state, Gate):  # pragma: no branch
            # Cannot be hit because of the semantics of
            # `needs_gate_evaluation_for_metadata_change`. Here to
            # appease mypy.
            raise RuntimeError(  # pragma: no cover
                "Label not in a gate",
            )

        could_progress = process_gate(
            app=app,
            state=current_state,
            state_machine=state_machine,
            label=label,
            conn=conn,
        )

    if could_progress:
        process_transitions(app, label)


def delete_label(app: App, label: LabelRef) -> None:
    """
    Deletes the metadata for a label and marks the label as deleted.

    The history for the label is not changed (in order to allow post-hoc
    analysis of the path the label took through the state machine).
    """
    state_machine = get_state_machine(app, label)  # Raises UnknownStateMachine

    with app.db.begin() as conn:
        try:
            row = lock_label(label, conn)
        except UnknownLabel:
            return

        if row is None or row.deleted:
            return

        # Record the label as having been deleted and remove its metadata
        conn.execute(labels.update().where(and_(
            labels.c.name == label.name,
            labels.c.state_machine == label.state_machine,
        )).values(
            metadata={},
            deleted=True,
        ))

        # Add a history entry for the deletion
        current_state = get_current_state(label, state_machine, conn)
        conn.execute(history.insert().values(
            label_name=label.name,
            label_state_machine=label.state_machine,
            old_state=current_state.name,
            new_state=None,
        ))


LabelStateProcessor = Callable[[App, State, StateMachine, LabelRef, Any], bool]


def process_cron(
    process: LabelStateProcessor,
    get_labels: Callable[[StateMachine, State, Any], Iterable[str]],
    app: App,
    state_machine: StateMachine,
    state: State,
):
    """
    Cron event entrypoint.
    """
    with app.db.begin() as conn:
        relevant_labels = get_labels(state_machine, state, conn)

    for label_name in relevant_labels:
        with suppress_exceptions(logger):
            label = LabelRef(name=label_name, state_machine=state_machine.name)
            could_progress = False

            with app.db.begin() as conn:
                lock_label(label, conn)
                current_state = get_current_state(label, state_machine, conn)

                if current_state != state:
                    continue

                could_progress = process(  # type: ignore
                    app=app,
                    state=state,
                    state_machine=state_machine,
                    label=label,
                    conn=conn,
                )

            if could_progress:
                process_transitions(app, label)



"""Utilities for state machine execution."""

import datetime
from typing import Any, Dict, List, Tuple, Iterable

import dateutil.tz
from sqlalchemy import and_, func, not_, select
from sqlalchemy.sql.elements import ClauseElement

from routemaster.db import labels, history
from routemaster.app import App
from routemaster.feeds import feeds_for_state_machine
from routemaster.config import Gate, State, StateMachine, ContextNextStates
from routemaster.context import Context
from routemaster.state_machine.types import LabelRef, Metadata
from routemaster.state_machine.exceptions import (
    UnknownLabel,
    UnknownStateMachine,
)


def get_state_machine(app: App, label: LabelRef) -> StateMachine:
    """Finds the state machine instance by name in the app config."""
    try:
        return app.config.state_machines[label.state_machine]
    except KeyError as k:
        raise UnknownStateMachine(label.state_machine)


def start_state_machine(app: App, label: LabelRef, conn) -> None:
    """Create the first history entry for a label in a state machine."""
    state_machine = get_state_machine(app, label)
    conn.execute(history.insert().values(
        label_name=label.name,
        label_state_machine=label.state_machine,
        old_state=None,
        new_state=state_machine.states[0].name,
    ))


def choose_next_state(
    state_machine: StateMachine,
    current_state: State,
    context: Context,
) -> State:
    """Assuming a transition is valid, choose a next state."""
    next_state_name = current_state.next_states.next_state_for_label(context)
    return state_machine.get_state(next_state_name)


def get_label_metadata(
    label: LabelRef,
    state_machine: StateMachine,
    conn,
) -> Tuple[Dict[str, Any], bool]:
    """Get the metadata and whether the label has been deleted."""
    return conn.execute(
        select([labels.c.metadata, labels.c.deleted]).where(and_(
            labels.c.name == label.name,
            labels.c.state_machine == state_machine.name,
        )),
    ).fetchone()


def get_current_state(
    label: LabelRef,
    state_machine: StateMachine,
    conn,
) -> State:
    """Get the current state of a label, based on its last history entry."""
    history_entry = conn.execute(
        select([history]).where(and_(
            history.c.label_name == label.name,
            history.c.label_state_machine == state_machine.name,
        )).order_by(
            history.c.id.desc(),
        ).limit(1),
    ).fetchone()

    if history_entry is None:
        raise UnknownLabel(label)

    return state_machine.get_state(history_entry.new_state)


def get_current_history(label: LabelRef, conn) -> Any:
    """Get a label's last history entry."""
    history_entry = conn.execute(
        select([history]).where(and_(
            history.c.label_name == label.name,
            history.c.label_state_machine == label.state_machine,
        )).order_by(
            history.c.id.desc(),
        ).limit(1),
    ).fetchone()

    if history_entry is None:
        raise UnknownLabel(label)

    return history_entry


def needs_gate_evaluation_for_metadata_change(
    state_machine: StateMachine,
    label: LabelRef,
    update: Metadata,
    conn,
) -> Tuple[bool, State]:
    """
    Given a change to the metadata, should the gate evaluation be triggered.
    """

    current_state = get_current_state(label, state_machine, conn)

    if not isinstance(current_state, Gate):
        # Label is not a gate state so there's no trigger to resolve.
        return False, current_state

    if any(
        trigger.should_trigger_for_update(update)
        for trigger in current_state.metadata_triggers
    ):
        return True, current_state

    return False, current_state


def lock_label(label: LabelRef, conn):
    """Lock a label in the current transaction."""
    row = conn.execute(
        labels.select().where(
            and_(
                labels.c.name == label.name,
                labels.c.state_machine == label.state_machine,
            ),
        ).with_for_update(),
    ).fetchone()

    if row is None:
        raise UnknownLabel(label)

    return row


def labels_in_state(
    state_machine: StateMachine,
    state: State,
    conn,
) -> List[str]:
    """Util to get all the labels in an action state that need retrying."""
    return _labels_in_state(
        state_machine,
        state,
        (),
        conn,
    )


def labels_needing_metadata_update_retry_in_gate(
    state_machine: StateMachine,
    state: State,
    conn,
) -> List[str]:
    """Util to get all the labels in an action state that need retrying."""
    if not isinstance(state, Gate):  # pragma: no branch
        raise ValueError(  # pragma: no cover
            f"labels_needing_metadata_update_retry_in_gate called with "
            f"{state.name} which is not an Gate",
        )

    return _labels_in_state(
        state_machine,
        state,
        (not_(labels.c.metadata_triggers_processed),),
        conn,
    )


def _labels_in_state(
    state_machine: StateMachine,
    state: State,
    filters: Iterable[ClauseElement],
    conn,
) -> List[str]:
    """Util to get all the labels in an action state that need retrying."""
    ranked_transitions = select((
        history.c.label_name,
        history.c.old_state,
        history.c.new_state,
        func.row_number().over(
            order_by=history.c.id.desc(),
            partition_by=history.c.label_name,
        ).label('rank'),
    )).where(
        history.c.label_state_machine == state_machine.name,
    ).alias('transitions')

    active_filters = (
        ranked_transitions.c.new_state == state.name,
        ranked_transitions.c.rank == 1,
        ranked_transitions.c.label_name == labels.c.name,
        labels.c.state_machine == state_machine.name,
    )

    active_participants = select((
        ranked_transitions.c.label_name,
    )).where(and_(*(active_filters + tuple(filters))))

    return [x for x, in conn.execute(active_participants)]


def context_for_label(
    label: LabelRef,
    metadata: Metadata,
    state_machine: StateMachine,
    state: State,
    history_entry: Any,
) -> Context:
    """Util to build the context for a label in a state."""
    feeds = feeds_for_state_machine(state_machine)

    accessed_variables: List[str] = []
    if isinstance(state, Gate):
        accessed_variables.extend(state.exit_condition.accessed_variables())
    if isinstance(state.next_states, ContextNextStates):
        accessed_variables.append(state.next_states.path)

    return Context(
        label=label.name,
        metadata=metadata,
        now=datetime.datetime.now(dateutil.tz.tzutc()),
        feeds=feeds,
        accessed_variables=accessed_variables,
        current_history_entry=history_entry,
    )



"""Public Database interface."""

from routemaster.db.model import labels, history, metadata
from routemaster.db.initialisation import initialise_db

__all__ = (
    'labels',
    'history',
    'metadata',
    'initialise_db',
)



"""Peephole evaluator optimiser."""

from routemaster.exit_conditions.operations import Operation

MATCHERS = [
    (
        [
            (Operation.NOT,),
            (Operation.NOT,),
        ],
        [
        ],
    ),
    (
        [
            (Operation.TO_BOOL,),
            (Operation.TO_BOOL,),
        ],
        [
            (Operation.TO_BOOL,),
        ],
    ),
    (
        [
            (Operation.NOT,),
            (Operation.TO_BOOL,),
        ],
        [
            (Operation.NOT,),
        ],
    ),
    (
        [
            (Operation.AND,),
            (Operation.TO_BOOL,),
        ],
        [
            (Operation.AND,),
        ],
    ),
    (
        [
            (Operation.OR,),
            (Operation.TO_BOOL,),
        ],
        [
            (Operation.OR,),
        ],
    ),
    (
        [
            (Operation.EQ,),
            (Operation.TO_BOOL,),
        ],
        [
            (Operation.EQ,),
        ],
    ),
    (
        [
            (Operation.LT,),
            (Operation.TO_BOOL,),
        ],
        [
            (Operation.LT,),
        ],
    ),
    (
        [
            (Operation.GT,),
            (Operation.TO_BOOL,),
        ],
        [
            (Operation.GT,),
        ],
    ),
]


def peephole_optimise(instructions):
    """Run peephole optimisations over a given instruction sequence."""
    instructions = list(instructions)

    any_changes = True

    while any_changes:
        any_changes = False

        for match_pattern, replace_pattern in MATCHERS:
            match_pattern_length = len(match_pattern)
            for index in range(len(instructions) + 1 - match_pattern_length):
                if (
                    instructions[index:index + match_pattern_length] ==
                    match_pattern
                ):
                    instructions[index:index + match_pattern_length] = \
                        replace_pattern
                    any_changes = True
                    break

        # TO_BOOL after property elimination
        for index in range(len(instructions) - 1):
            if (
                instructions[index][0] == Operation.PROPERTY and
                instructions[index + 1][0] == Operation.TO_BOOL
            ):
                del instructions[index + 1]
                any_changes = True
                break

    return instructions



"""Periodic job running."""

import time
import logging
import functools
import itertools
import threading
from typing import Any, List, Callable

import schedule

from routemaster.app import App
from routemaster.config import (
    Gate,
    State,
    Action,
    TimeTrigger,
    StateMachine,
    IntervalTrigger,
    MetadataTrigger,
)
from routemaster.state_machine import (
    LabelStateProcessor,
    process_cron,
    process_gate,
    process_action,
    labels_in_state,
    labels_needing_metadata_update_retry_in_gate,
)

logger = logging.getLogger(__name__)

IsTerminating = Callable[[], bool]

# Note: This function will be called in a different transaction to where we
# iterate over the results, so to prevent confusion or the possible
# introduction of errors, we require all the data up-front.
LabelProvider = Callable[[StateMachine, State, Any], List[str]]

CronProcessor = Callable[
    [LabelStateProcessor, LabelProvider, State, StateMachine],
    None,
]


# The cron configuration works by building up a partially applied function
# `process_job`, at multiple levels. This allows us to hook an intermediate
# stage for testing.

def process_job(
    *,
    # Bound at the cron thread level
    is_terminating: IsTerminating,
    # Bound at the state scheduling level
    app: App,
    state: State,
    state_machine: StateMachine,
    # Bound when scheduling a specific job for a state
    fn: LabelStateProcessor,
    label_provider: LabelProvider,
):
    """Process a single instance of a single cron job."""

    def _iter_labels_until_terminating(state_machine, state, conn):
        return itertools.takewhile(
            lambda _: not is_terminating(),
            label_provider(state_machine, state, conn),
        )

    logger.info(
        f"Started cron {fn.__name__} for state {state.name} in "
        f"{state_machine.name}",
    )

    try:
        time_start = time.time()
        process_cron(
            process=fn,
            get_labels=_iter_labels_until_terminating,
            app=app,
            state=state,
            state_machine=state_machine,
        )
        duration = time.time() - time_start
    except Exception:
        logger.exception(f"Error while processing cron {fn.__name__}")
        return

    logger.info(
        f"Completed cron {fn.__name__} for state {state.name} "
        f"in {state_machine.name} in {duration:.2f} seconds",
    )


def _configure_schedule_for_state(
    scheduler: schedule.Scheduler,
    processor: CronProcessor,
    state: State,
) -> None:
    if isinstance(state, Action):
        scheduler.every().minute.do(
            processor,
            fn=process_action,
            label_provider=labels_in_state,
        )
    elif isinstance(state, Gate):
        for trigger in state.triggers:
            if isinstance(trigger, TimeTrigger):
                scheduler.every().day.at(
                    f"{trigger.time.hour:02d}:{trigger.time.minute:02d}",
                ).do(
                    processor,
                    fn=process_gate,
                    label_provider=labels_in_state,
                )
            elif isinstance(trigger, IntervalTrigger):
                scheduler.every(
                    trigger.interval.total_seconds(),
                ).seconds.do(
                    processor,
                    fn=process_gate,
                    label_provider=labels_in_state,
                )
            elif isinstance(trigger, MetadataTrigger):  # pragma: no branch
                label_provider = labels_needing_metadata_update_retry_in_gate
                scheduler.every().minute.do(
                    processor,
                    fn=process_gate,
                    label_provider=label_provider,
                )
            else:
                # We only care about time based triggers and retries here.
                pass  # pragma: no cover
    else:
        raise RuntimeError(  # pragma: no cover
            f"Unsupported state type {state}",
        )


def configure_schedule(
    app: App,
    scheduler: schedule.Scheduler,
    processor: CronProcessor,
) -> None:
    """Set up all scheduled tasks that need running."""
    for state_machine in app.config.state_machines.values():
        for state in state_machine.states:
            _configure_schedule_for_state(
                scheduler,
                functools.partial(
                    processor,
                    app=app,
                    state=state,
                    state_machine=state_machine,
                ),
                state,
            )


class CronThread(threading.Thread):  # pragma: no cover
    """Background thread for running periodic jobs."""

    def __init__(self, app: App) -> None:
        self._terminating = False
        self.app = app
        self.scheduler = schedule.Scheduler()
        super().__init__(name="cron")

    def run(self) -> None:
        """Run main scheduling loop."""
        configure_schedule(
            self.app,
            self.scheduler,
            functools.partial(
                process_job,
                is_terminating=self.is_terminating,
            ),
        )
        logger.info("Starting cron thread")
        while not self.is_terminating():
            self.scheduler.run_pending()
            time.sleep(1)

    def stop(self) -> None:
        """Set the stopping flag and wait for thread end."""
        self._terminating = True
        logger.info("Cron thread shutting down")
        self.join()

    def is_terminating(self) -> bool:
        """Dynamically access whether we are terminating."""
        return self._terminating



"""Validation of state machines."""
import networkx
from sqlalchemy import and_, func, false, select

from routemaster.db import labels, history
from routemaster.app import App
from routemaster.config import Config, StateMachine


class ValidationError(Exception):
    """Class for errors raised to indicate invalid configuration."""
    pass


def validate_config(app: App, config: Config):
    """Validate that a given config satisfies invariants."""
    for state_machine in config.state_machines.values():
        _validate_state_machine(app, state_machine)


def _validate_state_machine(app: App, state_machine: StateMachine):
    """Validate that a given state machine is internally consistent."""
    _validate_route_start_to_end(state_machine)
    _validate_all_states_exist(state_machine)
    _validate_no_labels_in_nonexistent_states(state_machine, app)


def _build_graph(state_machine: StateMachine) -> networkx.Graph:
    graph = networkx.Graph()
    for state in state_machine.states:
        graph.add_node(state.name)
        for destination_name in state.next_states.all_destinations():
            graph.add_edge(state.name, destination_name)
    return graph


def _validate_route_start_to_end(state_machine):
    graph = _build_graph(state_machine)
    if not networkx.is_connected(graph):
        raise ValidationError("Graph is not fully connected")


def _validate_all_states_exist(state_machine):
    state_names = set(x.name for x in state_machine.states)
    for state in state_machine.states:
        for destination_name in state.next_states.all_destinations():
            if destination_name not in state_names:
                raise ValidationError(f"{destination_name} does not exist")


def _validate_no_labels_in_nonexistent_states(state_machine, app):
    states = [x.name for x in state_machine.states]

    with app.db.begin() as conn:

        latest_states = select([
            func.max(history.c.created).label('latest'),
            history.c.label_name,
            history.c.label_state_machine,
        ]).group_by(
            history.c.label_name,
            history.c.label_state_machine,
        ).alias('latest_states')

        invalid_states = select([
            history.c.new_state,
            func.count(history.c.id),
        ]).select_from(
            history.join(
                latest_states,
                and_(
                    latest_states.c.label_name == history.c.label_name,
                    latest_states.c.latest == history.c.created,
                    (
                        latest_states.c.label_state_machine ==
                        history.c.label_state_machine
                    ),
                ),
            ).join(
                labels,
                and_(
                    labels.c.name == history.c.label_name,
                    labels.c.state_machine == history.c.label_state_machine,
                ),
            ),
        ).where(
            and_(
                labels.c.deleted == false(),
                ~history.c.new_state.in_(states),
            ),
        ).group_by(
            history.c.label_name,
            history.c.label_state_machine,
            history.c.new_state,
        )

        result = conn.execute(invalid_states)
        inhabited = dict(result.fetchall())
        if inhabited:
            raise ValidationError(
                f"Labels currently in states that no longer exist: "
                f"{', '.join(inhabited)}",
            )



"""Core App singleton that holds state for the application."""
import threading

from sqlalchemy.engine import Engine

from routemaster.db import initialise_db
from routemaster.config import Config


class App(threading.local):
    """Core application state."""

    db: Engine
    config: Config

    def __init__(self, config: Config, log_level: str = 'INFO') -> None:
        """Initialisation of the app state."""
        self.config = config
        self.db = initialise_db(self.config.database)
        self.log_level = log_level



"""Global test setup and fixtures."""

import os
import re
import json
import datetime
import contextlib
from typing import Any, Dict

import mock
import pytest
import httpretty
import dateutil.tz
from sqlalchemy import and_, select, create_engine

from routemaster import state_machine
from routemaster.db import labels, history, metadata
from routemaster.app import App
from routemaster.utils import dict_merge
from routemaster.config import (
    Gate,
    Action,
    Config,
    Webhook,
    FeedConfig,
    NoNextStates,
    StateMachine,
    DatabaseConfig,
    OnEntryTrigger,
    MetadataTrigger,
    ConstantNextState,
    ContextNextStates,
    ContextNextStatesOption,
)
from routemaster.server import server
from routemaster.context import Context
from routemaster.webhooks import WebhookResult
from routemaster.state_machine import LabelRef
from routemaster.exit_conditions import ExitConditionProgram

TEST_DATABASE_CONFIG = DatabaseConfig(
    host=os.environ.get('PG_HOST', 'localhost'),
    port=int(os.environ.get('PG_PORT', 5432)),
    name=os.environ.get('PG_DB', 'routemaster_test'),
    username=os.environ.get('PG_USER', ''),
    password=os.environ.get('PG_PASS', ''),
)

TEST_STATE_MACHINES = {
    'test_machine': StateMachine(
        name='test_machine',
        feeds=[
            FeedConfig(name='tests', url='http://localhost/tests'),
        ],
        webhooks=[
            Webhook(
                match=re.compile('//(.+\\.)?example\\.com'),
                headers={
                    'x-api-key': 'Rahfew7eed1ierae0moa2sho3ieB1et3ohhum0Ei',
                },
            ),
        ],
        states=[
            Gate(
                name='start',
                triggers=[
                    MetadataTrigger(metadata_path='should_progress'),
                    OnEntryTrigger(),
                ],
                next_states=ContextNextStates(
                    path='feeds.tests.should_do_alternate_action',
                    destinations=[
                        ContextNextStatesOption(
                            state='perform_action',
                            value=False,
                        ),
                        ContextNextStatesOption(
                            state='perform_alternate_action',
                            value=True,
                        ),
                    ],
                    default='perform_action',
                ),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Action(
                name='perform_action',
                webhook='http://localhost/hook/<state_machine>/<label>',
                next_states=ConstantNextState(state='end'),
            ),
            Action(
                name='perform_alternate_action',
                webhook='http://localhost/hook',
                next_states=ContextNextStates(
                    path='feeds.tests.should_loop',
                    destinations=[
                        ContextNextStatesOption(
                            state='end',
                            value=False,
                        ),
                        ContextNextStatesOption(
                            state='start',
                            value=True,
                        ),
                    ],
                    default='end',
                ),
            ),
            Gate(
                name='end',
                triggers=[],
                next_states=NoNextStates(),
                exit_condition=ExitConditionProgram('false'),
            ),
        ],
    ),
    # This state machine is used for exercising race conditions in tests and is
    # purposefully not realistic.
    'test_machine_2': StateMachine(
        name='test_machine_2',
        feeds=[],
        webhooks=[],
        states=[
            Gate(
                name='gate_1',
                triggers=[
                    OnEntryTrigger(),
                    MetadataTrigger(metadata_path='should_progress'),
                ],
                next_states=ConstantNextState('gate_2'),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Gate(
                name='gate_2',
                triggers=[
                    MetadataTrigger(metadata_path='should_progress'),
                ],
                next_states=ConstantNextState('end'),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Gate(
                name='end',
                triggers=[],
                next_states=NoNextStates(),
                exit_condition=ExitConditionProgram('false'),
            ),
        ],
    ),
    # This state machine is used for testing the possibility of infinite loops.
    'test_infinite_machine': StateMachine(
        name='test_infinite_machine',
        feeds=[],
        webhooks=[],
        states=[
            Gate(
                name='gate_1',
                triggers=[
                    OnEntryTrigger(),
                    MetadataTrigger(metadata_path='should_progress'),
                ],
                next_states=ConstantNextState('gate_2'),
                exit_condition=ExitConditionProgram(
                    'metadata.should_progress = true',
                ),
            ),
            Gate(
                name='gate_2',
                triggers=[OnEntryTrigger()],
                next_states=ConstantNextState('gate_3'),
                exit_condition=ExitConditionProgram('true'),
            ),
            Gate(
                name='gate_3',
                triggers=[OnEntryTrigger()],
                next_states=ConstantNextState('gate_2'),
                exit_condition=ExitConditionProgram('true'),
            ),
        ],
    ),
    'test_machine_timing': StateMachine(
        name='test_machine_timing',
        feeds=[],
        webhooks=[],
        states=[
            Gate(
                name='start',
                triggers=[
                    OnEntryTrigger(),
                ],
                next_states=ConstantNextState('end'),
                exit_condition=ExitConditionProgram(
                    '1d12h has passed since history.entered_state',
                ),
            ),
            Gate(
                name='end',
                triggers=[],
                next_states=NoNextStates(),
                exit_condition=ExitConditionProgram('false'),
            ),
        ],
    ),
}

TEST_ENGINE = create_engine(TEST_DATABASE_CONFIG.connstr)


class TestApp(App):
    """
    Mocked App subclass which overloads the `db` property.

    This has two implications:

    1. We use a shared engine in all the tests rather than connecting many
       times (a speed improvement),
    2. We can set a flag on access to `.db` so that we needn't bother with
       resetting the database if nothing has actually been changed.
    """
    def __init__(self, config):
        self.config = config
        self.database_used = False

    @property
    def db(self):
        """Get the shared DB and set the used flag."""
        self.database_used = True
        return TEST_ENGINE


@pytest.fixture()
def app(**kwargs):
    """Create the Flask app for testing."""
    server.config.app = app_config(**kwargs)
    return server


@pytest.fixture()
def app_config(**kwargs):
    """Create an `App` config object for testing."""
    return TestApp(Config(
        state_machines=kwargs.get('state_machines', TEST_STATE_MACHINES),
        database=kwargs.get('database', TEST_DATABASE_CONFIG),
    ))


@pytest.fixture()
def custom_app_config():
    """Return the app config fixture directly so that we can modify config."""
    return app_config


@pytest.fixture()
def app_env():
    """
    Create a dict of environment variables.

    Mirrors the testing environment for use in subprocesses.
    """
    return {
        'DB_HOST': TEST_DATABASE_CONFIG.host,
        'DB_PORT': str(TEST_DATABASE_CONFIG.port),
        'DB_NAME': TEST_DATABASE_CONFIG.name,
        'DB_USER': TEST_DATABASE_CONFIG.username,
        'DB_PASS': TEST_DATABASE_CONFIG.password,
    }


@pytest.fixture(autouse=True, scope='session')
def database_creation(request):
    """Wrap test session in creating and destroying all required tables."""
    metadata.create_all(bind=TEST_ENGINE)
    request.addfinalizer(lambda: metadata.drop_all(bind=TEST_ENGINE))


@pytest.yield_fixture(autouse=True)
def database_clear(app_config):
    """Truncate all tables after each test."""
    yield
    if app_config.database_used:
        with app_config.db.begin() as conn:
            for table in metadata.tables:
                conn.execute(f'truncate table {table} cascade')


@pytest.fixture()
def create_label(app_config, mock_test_feed):
    """Create a label in the database."""

    def _create(
        name: str,
        state_machine_name: str,
        metadata: Dict[str, Any],
    ) -> LabelRef:
        with mock_test_feed():
            state_machine.create_label(
                app_config,
                LabelRef(name, state_machine_name),
                metadata,
            )
            return LabelRef(name, state_machine_name)

    return _create


@pytest.fixture()
def delete_label(app_config):
    """
    Mark a label in the database as deleted.
    """

    def _delete(name: str, state_machine_name: str) -> None:
        state_machine.delete_label(
            app_config,
            LabelRef(name, state_machine_name),
        )

    return _delete


@pytest.fixture()
def create_deleted_label(create_label, delete_label):
    """
    Create a label in the database and then delete it.
    """

    def _create_and_delete(name: str, state_machine_name: str) -> LabelRef:
        create_label(name, state_machine_name, {})
        delete_label(name, state_machine_name)
        return LabelRef(name, state_machine_name)

    return _create_and_delete


@pytest.fixture()
def mock_webhook():
    """Mock the test config's webhook call."""
    @contextlib.contextmanager
    def _mock(result=WebhookResult.SUCCESS):
        runner = mock.Mock(return_value=result)
        with mock.patch(
            'routemaster.webhooks.RequestsWebhookRunner',
            return_value=runner,
        ):
            yield runner
    return _mock


@pytest.fixture()
def mock_test_feed():
    """Mock out the test feed."""
    @contextlib.contextmanager
    def _mock(data={'should_do_alternate_action': False}):
        httpretty.enable()
        httpretty.register_uri(
            httpretty.GET,
            'http://localhost/tests',
            body=json.dumps(data),
            content_type='application/json',
        )

        try:
            yield
        finally:
            httpretty.disable()
            httpretty.reset()

    return _mock


@pytest.fixture()
def assert_history(app_config):
    """Assert that the database history matches what is expected."""
    def _assert(entries):
        with app_config.db.begin() as conn:
            history_entries = [
                tuple(x)
                for x in conn.execute(
                    select((
                        history.c.old_state,
                        history.c.new_state,
                    )).order_by(history.c.id.asc()),
                )
            ]

            assert history_entries == entries
    return _assert


@pytest.fixture()
def set_metadata(app_config):
    """Directly set the metadata for a label in the database."""
    def _inner(label, update):
        with app_config.db.begin() as conn:
            filter_ = and_(
                labels.c.name == label.name,
                labels.c.state_machine == label.state_machine,
            )

            existing_metadata = conn.scalar(
                select([labels.c.metadata]).where(filter_),
            )

            new_metadata = dict_merge(existing_metadata, update)

            conn.execute(labels.update().where(filter_).values(
                metadata=new_metadata,
                metadata_triggers_processed=True,
            ))

            return new_metadata
    return _inner


@pytest.fixture()
def make_context():
    """Factory for Contexts that provides sane defaults for testing."""
    def _inner(**kwargs):
        return Context(
            label=kwargs['label'],
            metadata=kwargs.get('metadata', {}),
            now=kwargs.get('now', datetime.datetime.now(dateutil.tz.tzutc())),
            feeds=kwargs.get('feeds', {}),
            accessed_variables=kwargs.get('accessed_variables', []),
            current_history_entry=kwargs.get('current_history_entry'),
        )
    return _inner
